package com.github.libretube.util

import android.util.Log
import com.github.libretube.api.RetrofitInstance
import com.github.libretube.api.obj.ContentItem
import com.github.libretube.api.obj.DeArrowContent
import com.github.libretube.api.obj.StreamItem
import com.github.libretube.api.obj.Streams
import com.github.libretube.constants.PreferenceKeys
import com.github.libretube.extensions.toID
import com.github.libretube.helpers.PreferenceHelper
import java.util.TreeSet

object DeArrowUtil {
    private fun extractTitleAndThumbnail(content: DeArrowContent): Pair<String?, String?> {
        val newTitle = content.titles.firstOrNull { it.votes >= 0 || it.locked }?.title
        val newThumbnail = content.thumbnails.firstOrNull {
            it.thumbnail != null && !it.original && (it.votes >= 0 || it.locked)
        }?.thumbnail
        return newTitle to newThumbnail
    }

    private suspend fun fetchDeArrowContent(videoIds: List<String>): Map<String, DeArrowContent>? {
        val videoIdsString = videoIds.mapTo(TreeSet()) { it }.joinToString(",")

        return try {
            RetrofitInstance.api.getDeArrowContent(videoIdsString)
        } catch (e: Exception) {
            Log.e(this::class.java.name, e.toString())
            null
        }
    }

    /**
     * Apply the new titles and thumbnails generated by DeArrow to the streams item
     */
    suspend fun deArrowStreams(streams: Streams, vidId: String): Streams {
        if (!PreferenceHelper.getBoolean(PreferenceKeys.DEARROW, false)) return streams

        val videoIds = listOf(vidId) + streams.relatedStreams.map { it.url!!.toID() }
        val response = fetchDeArrowContent(videoIds) ?: return streams

        for ((videoId, data) in response.entries) {
            val (newTitle, newThumbnail) = extractTitleAndThumbnail(data)

            if (videoId == vidId) {
                if (newTitle != null) streams.title = newTitle
                if (newThumbnail != null) streams.thumbnailUrl = newThumbnail
            } else {
                val streamItem = streams.relatedStreams
                    .firstOrNull { it.url?.toID() == videoId } ?: continue

                if (newTitle != null) streamItem.title = newTitle
                if (newThumbnail != null) streamItem.thumbnail = newThumbnail
            }
        }

        return streams
    }

    /**
     * Apply the new titles and thumbnails generated by DeArrow to the stream items
     */
    suspend fun deArrowStreamItems(streamItems: List<StreamItem>): List<StreamItem> {
        if (!PreferenceHelper.getBoolean(PreferenceKeys.DEARROW, false)) return streamItems

        val response = fetchDeArrowContent(streamItems.map{ it.url!!.toID() }) ?: return streamItems

        for ((videoId, data) in response.entries) {
            val (newTitle, newThumbnail) = extractTitleAndThumbnail(data)
            val streamItem = streamItems.firstOrNull { it.url?.toID() == videoId } ?: continue

            if (newTitle != null) streamItem.title = newTitle
            if (newThumbnail != null) streamItem.thumbnail = newThumbnail
        }
        return streamItems
    }

    /**
     * Apply the new titles and thumbnails generated by DeArrow to the stream items
     */
    suspend fun deArrowContentItems(contentItems: List<ContentItem>): List<ContentItem> {
        if (!PreferenceHelper.getBoolean(PreferenceKeys.DEARROW, false)) return contentItems

        val videoIds = contentItems.filter { it.type == "stream" }
            .map { it.url.toID() }

        if (videoIds.isEmpty()) return contentItems

        val response = fetchDeArrowContent(videoIds) ?: return contentItems

        for ((videoId, data) in response.entries) {
            val (newTitle, newThumbnail) = extractTitleAndThumbnail(data)
            val contentItem = contentItems.firstOrNull { it.url.toID() == videoId } ?: continue

            if (newTitle != null) { contentItem.title = newTitle }
            if (newThumbnail != null) { contentItem.thumbnail = newThumbnail }
        }
        return contentItems
    }
}

/**
 * If enabled in the preferences, this overrides the video's thumbnail and title with the one
 * provided by the DeArrow project
 */
@JvmName("deArrowStreamItems")
suspend fun List<StreamItem>.deArrow() = DeArrowUtil.deArrowStreamItems(this)

/**
 * If enabled in the preferences, this overrides the video's thumbnail and title with the one
 * provided by the DeArrow project
 */
@JvmName("deArrowContentItems")
suspend fun List<ContentItem>.deArrow() = DeArrowUtil.deArrowContentItems(this)

@JvmName("deArrowStreams")
suspend fun Streams.deArrow(videoId: String) = DeArrowUtil.deArrowStreams(this, videoId)